#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2021 All Rights Reserved.
   ------------------------------------------------------
   File Name : test_homo
   Author : qizhi.zqz
   Email: qizhi.zqz@alibaba-inc.com
   Create Time : 2021/9/16 下午9:14
   Description : description what the main function of this file
"""

import tensorflow as tf
import numpy as np
tf.compat.v1.disable_eager_execution()

size=100
homo_module = tf.load_op_library("/Users/qizhi.zqz/projects/morse-stf/morse-stf/cops/_stf_homo_macos.so")
random_module = tf.load_op_library("/Users/qizhi.zqz/projects/morse-stf/morse-stf/cops/_stf_random_macos.so")

def test_mat_mul_vec():
    sk, pk, gk = homo_module.gen_key(tf.Variable(initial_value=[0])) # 一定要以这个tf.Variable作为入参，否则会有错误
    row_num = 300
    col_num = 100
    plaintext = random_module.rint64(shape=[col_num], step=tf.Variable(initial_value=[0]))
    #plaintext = tf.cast(tf.range(start=0, limit=size), 'int64') # (shape=[size], dtype='int64')

    cipher = homo_module.enc(pk, plaintext)
    print(cipher)

    #A = tf.constant(np.random.random_integers(-(1<<63), (1<<63)-1, size=[size,size]), dtype='int64')
    A=tf.random.uniform(shape=[row_num, col_num], minval=-(1<<63), maxval=(1<<63)-1, dtype='int64')

    plaintext_out=tf.squeeze(tf.matmul(A, tf.expand_dims(plaintext, axis=-1)))

    cipher_v_out = homo_module.mat_mul_vec(pk, gk, A, cipher)
    print("cipher_v_out=", cipher_v_out)



    plaintext_out2 = homo_module.dec(sk, [row_num], cipher_v_out)
    print(plaintext_out2)

    sess=tf.compat.v1.Session()

    sess.run(tf.compat.v1.initialize_all_variables())

    print(np.sum((sess.run(plaintext_out - plaintext_out2))**2))
    #sess=tf.compat.v1.Session()
    #print(sess.run(keys))

def test_mat_mul_vec_to_share():
    sk, pk, gk = homo_module.gen_key(tf.Variable(initial_value=[0]))  # 一定要以这个tf.Variable作为入参，否则会有错误
    row_num = 3000
    col_num = 128
    #plaintext = random_module.rint64(shape=[col_num], step=tf.Variable(initial_value=[0]))
    plaintext = tf.cast(tf.range(start=0, limit=col_num), 'int64') # (shape=[size], dtype='int64')

    cipher = homo_module.enc(pk, plaintext)
    print(cipher)

    # A = tf.constant(np.random.random_integers(-(1<<63), (1<<63)-1, size=[size,size]), dtype='int64')
    A = tf.random.uniform(shape=[row_num, col_num], minval=-(1 << 63), maxval=(1 << 63) - 1, dtype='int64')

    plaintext_out = tf.squeeze(tf.matmul(A, tf.expand_dims(plaintext, axis=-1)))

    plaintext_out1, cipher_v_out = homo_module.mat_mul_vec_to_share(pk, gk, A, cipher)
    print("cipher_v_out=", cipher_v_out)

    plaintext_out2 = homo_module.dec(sk, [row_num], cipher_v_out)
    print(plaintext_out2)

    sess = tf.compat.v1.Session()

    sess.run(tf.compat.v1.initialize_all_variables())

    print(np.sum((sess.run(plaintext_out - (plaintext_out1+plaintext_out2) )) ** 2))


def test_mat_mul_mat_to_share():
    sk, pk, gk = homo_module.gen_key(tf.Variable(initial_value=[0]))  # 一定要以这个tf.Variable作为入参，否则会有错误

    plaintext = random_module.rint64(shape=[2,3,4, size,1], step=tf.Variable(initial_value=[0]))
    # plaintext = tf.cast(tf.range(start=0, limit=size), 'int64') # (shape=[size], dtype='int64')
    cipher = homo_module.enc(pk, plaintext)
    print(cipher)

    # A = tf.constant(np.random.random_integers(-(1<<63), (1<<63)-1, size=[size,size]), dtype='int64')
    A = tf.random.uniform(shape=[2,3,4, size, size], minval=-(1 << 63), maxval=(1 << 63) - 1, dtype='int64')

    plaintext_out = tf.squeeze(tf.matmul(A, tf.expand_dims(plaintext, axis=-1)))

    plaintext_out1, cipher_v_out = homo_module.mat_mul_vec_to_share(pk, gk, A, cipher)
    print("cipher_v_out=", cipher_v_out)

    plaintext_out2 = homo_module.dec(sk, [size], cipher_v_out)
    print(plaintext_out2)

    sess = tf.compat.v1.Session()

    sess.run(tf.compat.v1.initialize_all_variables())

    print(np.sum((sess.run(plaintext_out - (plaintext_out1+plaintext_out2) )) ** 2))


def test_vec_mul_vec():
    sk, pk, gk = homo_module.gen_key(tf.Variable(initial_value=[0])) # 一定要以这个tf.Variable作为入参，否则会有错误
    col_num = 100
    plaintext = random_module.rint64(shape=[col_num], step=tf.Variable(initial_value=[0]))
    #plaintext = tf.cast(tf.range(start=0, limit=size), 'int64') # (shape=[size], dtype='int64')

    cipher = homo_module.enc(pk, plaintext)
    print(cipher)

    #A = tf.constant(np.random.random_integers(-(1<<63), (1<<63)-1, size=[size,size]), dtype='int64')
    A=tf.random.uniform(shape=[col_num], minval=-(1<<63), maxval=(1<<63)-1, dtype='int64')

    plaintext_out=A * plaintext

    cipher_v_out = homo_module.vec_mul_vec(pk, A, cipher)
    print("cipher_v_out=", cipher_v_out)



    plaintext_out2 = homo_module.dec(sk, [col_num], cipher_v_out)
    print(plaintext_out2)

    sess=tf.compat.v1.Session()

    sess.run(tf.compat.v1.initialize_all_variables())

    print(np.sum((sess.run(plaintext_out - plaintext_out2))**2))
    #sess=tf.compat.v1.Session()
    #print(sess.run(keys))





def test_cipher_mul_share():
    sk, pk, gk = homo_module.gen_key(tf.Variable(initial_value=[0])) # 一定要以这个tf.Variable作为入参，否则会有错误
    col_num = 100
    plaintext = random_module.rint64(shape=[col_num], step=tf.Variable(initial_value=[0]))
    #plaintext = tf.cast(tf.range(start=0, limit=size), 'int64') # (shape=[size], dtype='int64')

    cipher = homo_module.enc(pk, plaintext)
    print(cipher)

    cipher_v_out, share_v_out = homo_module.cipher_to_share(col_num, pk, cipher)
    print("cipher_v_out=", cipher_v_out)
    print("share_v_out=", share_v_out)

    share_v_out2 = homo_module.dec(sk, col_num, cipher_v_out)
    sess=tf.compat.v1.Session()

    sess.run(tf.compat.v1.initialize_all_variables())
    #print(sess.run([cipher_v_out, share_v_out]))
    print(np.sum((sess.run(share_v_out + share_v_out2 - plaintext))**2))



if __name__ == '__main__':
    test_mat_mul_vec()
    # test_vec_mul_vec()
    # test_cipher_mul_share()
    # test_mat_mul_vec_to_share()


#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : ReLU
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-09-14 17:36
   Description : description what the main function of this file
"""

from stensorflow.ml.nn.layers.layer import Layer
from stensorflow.basic.basic_class.private import PrivateTensor, PrivateTensorBase
from stensorflow.basic.basic_class.pair import SharedVariablePair, SharedPair
from stensorflow.basic.operator.relu import relu, drelu_binary, relu_pull_back, relu_local, drelu_local
from typing import Union, List, Dict



class ReLU_bak(Layer):
    def __init__(self, output_dim: int, fathers: List[Layer]):
        if len(fathers) != 1:
            raise Exception("must have len(fathers) == 1 ")
        if fathers[0].output_dim != output_dim:
            raise Exception("must have fathers[0].output_dim == output_dim")
        super(ReLU, self).__init__(output_dim=output_dim, fathers=fathers)

    def func(self, w: List[SharedVariablePair], x: List[Union[PrivateTensor, SharedPair]]):
        if len(x) != 1:
            raise Exception("must have len(x) == 1")
        return relu(x[0])

    def pull_back(self, w: List[SharedPair], x: List[Union[PrivateTensor, SharedPair]], y: SharedPair,
                  ploss_py: SharedPair) \
            -> (List[SharedPair], Dict[Layer, SharedPair]):
        if len(x) != 1:
            raise Exception("must have len(x)==1")
        # ploss_px = {self.fathers[0]: selectshare(is_positive(x[0]), ploss_py)}
        ploss_px = {self.fathers[0]: relu_pull_back(x[0], ploss_py)}
        return [], ploss_px




class ReLU(Layer):
    def __init__(self, output_dim: int, fathers: List[Layer]):
        if len(fathers) != 1:
            raise Exception("must have len(fathers) == 1 ")
        if fathers[0].output_dim != output_dim:
            raise Exception("must have fathers[0].output_dim == output_dim")
        super(ReLU, self).__init__(output_dim=output_dim, fathers=fathers)
        self.drelu_b = None

    def func(self, w: List[SharedVariablePair], x: List[Union[PrivateTensor, SharedPair]]):
        if len(x) != 1:
            raise Exception("must have len(x) == 1")
        self.drelu_b = drelu_binary(x[0])
        return relu(x[0], drelu_b=self.drelu_b)

    def pull_back(self, w: List[SharedPair], x: List[Union[PrivateTensor, SharedPair]], y: SharedPair,
                  ploss_py: SharedPair) \
            -> (List[SharedPair], Dict[Layer, SharedPair]):
        if len(x) != 1:
            raise Exception("must have len(x)==1")
        # ploss_px = {self.fathers[0]: selectshare(is_positive(x[0]), ploss_py)}
        ploss_px = {self.fathers[0]: relu_pull_back(x[0], ploss_py, self.drelu_b)}
        return [], ploss_px


class ReLU_Local(Layer):
    def __init__(self, output_dim: int, fathers: List[Layer], owner):
        if len(fathers) != 1:
            raise Exception("must have len(fathers) == 1 ")
        if fathers[0].output_dim != output_dim:
            raise Exception("must have fathers[0].output_dim == output_dim")
        self.owner = owner
        super(ReLU_Local, self).__init__(output_dim=output_dim, fathers=fathers)

    def func(self, w, x: List[Union[PrivateTensor, SharedPair]]):
        if len(x) != 1:
            raise Exception("must have len(x) == 1")
        if not isinstance(x[0], PrivateTensorBase):
            x = x[0].to_private(owner=self.owner)
        else:
            x = x[0]
        return relu_local(x)

    def pull_back(self, w: List[SharedPair], x: List[Union[PrivateTensor, SharedPair]], y: PrivateTensor,
                  ploss_py: Union[PrivateTensor, SharedPair]) \
            -> (List[PrivateTensor], Dict[Layer, PrivateTensor]):
        if len(x) != 1:
            raise Exception("must have len(x)==1")
        if not isinstance(x[0], PrivateTensorBase):
            x = x[0].to_private(owner=self.owner)
        else:
            x = x[0]
        ploss_py = ploss_py.to_private(owner=self.owner)
        ploss_px = drelu_local(x) * ploss_py
        ploss_px = {self.fathers[0]: ploss_px}
        return [], ploss_px



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2021 All Rights Reserved.
   ------------------------------------------------------
   File Name : exception
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2021/5/19 下午3:51
   Description : description what the main function of this file
"""
import warnings


class StfException(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __repr__(self):
        return self.msg

    def __str__(self):
        return self.__repr__()

    def __unicode__(self):
        return self.__repr__()


class StfNoneException(StfException):
    def __init__(self, obj_name):
        self.msg = "{} should not be None.".format(obj_name)


class StfTypeException(StfException):
    def __init__(self, obj_name, expect_type, real_type):
        self.msg = "{} should be {}, but is {}.".format(obj_name, expect_type, real_type)


class StfDTypeException(StfException):
    def __init__(self, obj_name, expect_dtypes, real_dtype):
        self.msg = "{}.dtype should in {}, but is {}.".format(obj_name, expect_dtypes, real_dtype)


class StfValueException(StfException):
    def __init__(self, var_name, expect_value, real_value):
        self.msg = "{} should equal to {}, but in fact equal to {}.".format(var_name, expect_value, real_value)


class StfEqualException(StfException):

    def __init__(self, var1_name, var2_name, var1, var2):
        self.msg = "should have {}={}, but in fact {}={}, {}={}".format(var1_name, var2_name, var1_name,
                                                                        var1, var2_name, var2)


class StfEqualWarning:
    def __init__(self, var1_name, var2_name, var1, var2):
        self.msg = "should have {}={}, but in fact {}={}, {}={}".format(var1_name, var2_name, var1_name,
                                                                        var1, var2_name, var2)
        warnings.warn(self.msg)


class StfCondException(StfException):
    def __init__(self, cond, real):
        self.msg = "must have {}, but in fact {}.".format(cond, real)


def check_is_not_None(**kwargs):
    dict_name_value = dict(kwargs)
    for name, value in dict_name_value.items():
        if value is None:
            raise StfNoneException(obj_name=name)



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : start_server
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-05-15 10:10
   Description : description what the main function of this file
"""
import tensorflow as tf
from stensorflow.global_var import StfConfig
import argparse

"""
cmd line:
export PYTHONPATH={your path}/morse-stf
"""

tf.compat.v1.disable_eager_execution()


def start_local_server(config_file):
    if StfConfig.serverL is None:
        StfConfig.load_config(config_file=config_file)
        if StfConfig.parties != 2 and StfConfig.parties != 3:
            raise Exception("must have StfConfig.parties == 2 or StfConfig.parties == 3")
        if StfConfig.parties == 3:
            cluster = tf.train.ClusterSpec({"workerL": StfConfig.workerL_hosts, "workerR": StfConfig.workerR_hosts, "RS": StfConfig.RS_hosts})
        else:
            cluster = tf.train.ClusterSpec({"workerL": StfConfig.workerL_hosts, "workerR": StfConfig.workerR_hosts})
        StfConfig.serverL = tf.distribute.Server(cluster, job_name="workerL", task_index=0)
        StfConfig.serverR = tf.distribute.Server(cluster, job_name="workerR", task_index=0)
        if StfConfig.parties == 3:
            StfConfig.serverRS = tf.distribute.Server(cluster, job_name="RS", task_index=0)



def start_distributed_server(config_file, job_name):
    StfConfig.load_config(config_file=config_file, job_name=job_name)
    if StfConfig.parties != 2 and StfConfig.parties != 3:
        raise Exception("must have StfConfig.parties == 2 or StfConfig.parties == 2")
    if StfConfig.parties == 3:
        cluster = tf.train.ClusterSpec({"workerL": StfConfig.workerL_hosts, "workerR": StfConfig.workerR_hosts, "RS": StfConfig.RS_hosts})
    else:
        cluster = tf.train.ClusterSpec({"workerL": StfConfig.workerL_hosts, "workerR": StfConfig.workerR_hosts})
    server = tf.distribute.Server(cluster, job_name=job_name, task_index=0)
    server.join()

def start_client(config_file, job_name):
    StfConfig.load_config(config_file=config_file, job_name=job_name)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--player', type=str, choices=["workerL", "workerR", "RS"], required=True)
    parser.add_argument('--config_file', type=str, default="./conf/config.json")
    args = parser.parse_args()
    job_name = args.player
    config_file = args.config_file

    start_distributed_server(config_file, job_name)


if __name__ == '__main__':
    main()






#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group Copyright (c) 2004-2021 All Rights Reserved.
"""
from stensorflow.global_var import StfConfig
from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase, SharedTensorBase
from typing import Union
from stensorflow.basic.protocol.bilinear_triangle import BiliinearTriangle
from stensorflow.basic.protocol.bilinear_map import BM_PrivateTensor_SharedPair, BM_PrivateTensor_PrivateTensor, \
    BM_SharedPair_SharedPair, BM_SharedPair_PrivateTensor




def conv2d(input: Union[PrivateTensorBase, SharedPairBase],
           filter: Union[PrivateTensorBase, SharedPairBase],
           strides,
           padding,
           data_format='NHWC',
           dilations=None,
           fixed_point=None,
           prf_flag=None) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    """
    See tf.conv2d
    """
    model = StfConfig.conv_module
    if fixed_point is None:
        fixed_point = max(input.fixedpoint, filter.fixedpoint)
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    input_is_private = isinstance(input, PrivateTensorBase)
    input_is_pair = isinstance(input, SharedPairBase)
    filters_is_private = isinstance(filter, PrivateTensorBase)
    filters_is_pair = isinstance(filter, SharedPairBase)

    def func_conv2d(x, y):
        return model.int64_conv2d(input=x, filter=y, strides=strides, padding=padding,
                                  data_format=data_format, dilations=dilations)

    def lambda_func(x, y):
        return SharedTensorBase(func_conv2d(x.inner_value, y.inner_value))

    if input_is_private and filters_is_private:
        if input.owner == filter.owner:
            result = func_conv2d(x=input.inner_value, y=filter.inner_value)
            result = PrivateTensorBase(inner_value=result, fixedpoint=input.fixedpoint + filter.fixedpoint,
                                       owner=input.owner)
        else:
            result = BM_PrivateTensor_PrivateTensor(input, filter, lambda_func, prf_flag=prf_flag)
    elif input_is_private and filters_is_pair:
        result = BM_PrivateTensor_SharedPair(input, filter, lambda_func, prf_flag=prf_flag)
    elif input_is_pair and filters_is_private:
        result = BM_SharedPair_PrivateTensor(input, filter, lambda_func, prf_flag=prf_flag)
    elif input_is_pair and filters_is_pair:
        result = BM_SharedPair_SharedPair(input, filter, lambda_func, prf_flag=prf_flag)
    else:
        raise Exception("type exception for type(x)={}, type(y)={}".format(type(input), type(filter)))
    return result.dup_with_precision(fixed_point)


def conv2d_backprop_input(input_sizes,
                          filter: Union[PrivateTensorBase, SharedPairBase],
                          out_backprop: Union[PrivateTensorBase, SharedPairBase],
                          strides=None,
                          padding=None,
                          data_format="NHWC",
                          dilations=[1, 1, 1, 1],
                          fixed_point=None,
                          prf_flag=None
                          ) -> Union[PrivateTensorBase, SharedPairBase]:
    """
    See  tf.conv2d_backprop_input
    """
    model = StfConfig.conv_module
    if fixed_point is None:
        fixed_point = max(filter.fixedpoint, out_backprop.fixedpoint)
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    # whether an object is an instance of a class
    filters_is_private = isinstance(filter, PrivateTensorBase)
    filters_is_pair = isinstance(filter, SharedPairBase)
    out_backprop_is_private = isinstance(out_backprop, PrivateTensorBase)
    out_backprop_is_pair = isinstance(out_backprop, SharedPairBase)

    def func_back_input(f, o):
        return model.int64_conv2d_backprop_input(input_sizes=input_sizes, filter=f,
                                                 out_backprop=o, strides=strides,
                                                 padding=padding, data_format=data_format,
                                                 dilations=dilations)

    def lambda_func(f, o):
        return SharedTensorBase(func_back_input(f.inner_value, o.inner_value))

    if filters_is_private and out_backprop_is_private:
        if filter.owner == out_backprop.owner:
            result = func_back_input(filter.inner_value, out_backprop.inner_value)
            result = PrivateTensorBase(inner_value=result, fixedpoint=filter.fixedpoint + out_backprop.fixedpoint,
                                       owner=filter.owner)
        else:
            result = BM_PrivateTensor_PrivateTensor(filter, out_backprop, lambda_func, prf_flag=prf_flag)
    elif filters_is_private and out_backprop_is_pair:
        result = BM_PrivateTensor_SharedPair(filter, out_backprop, lambda_func, prf_flag=prf_flag)
    elif filters_is_pair and out_backprop_is_private:
        result = BM_SharedPair_PrivateTensor(filter, out_backprop, lambda_func, prf_flag=prf_flag)
    elif filters_is_pair and out_backprop_is_pair:
        result = BM_SharedPair_SharedPair(filter, out_backprop, lambda_func, prf_flag=prf_flag)
    else:
        raise Exception("type exception for type(x)={}, type(y)={}".format(type(filter), type(out_backprop)))
    return result.dup_with_precision(fixed_point)


def conv2d_backprop_filter(input: Union[PrivateTensorBase, SharedPairBase],
                           filter_sizes,
                           out_backprop: Union[PrivateTensorBase, SharedPairBase],
                           strides,
                           padding,
                           data_format="NHWC",
                           dilations=[1, 1, 1, 1],
                           fixed_point=None,
                           prf_flag=None
                           ) -> Union[PrivateTensorBase, SharedPairBase]:
    """
    See conv2d_backprop_filter
    """
    model = StfConfig.conv_module
    if fixed_point is None:
        fixed_point = max(input.fixedpoint, out_backprop.fixedpoint)
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    # whether an object is an instance of a class
    input_is_private = isinstance(input, PrivateTensorBase)
    input_is_pair = isinstance(input, SharedPairBase)
    out_backprop_is_private = isinstance(out_backprop, PrivateTensorBase)
    out_backprop_is_pair = isinstance(out_backprop, SharedPairBase)

    def func_back_filter(x, o):
        return model.int64_conv2d_backprop_filter(input=x,
                                                  filter_sizes=filter_sizes,
                                                  out_backprop=o,
                                                  strides=strides,
                                                  padding=padding,
                                                  data_format=data_format,
                                                  dilations=dilations)

    def lambda_func(x, o):
        return SharedTensorBase(func_back_filter(x.inner_value, o.inner_value))

    if input_is_private and out_backprop_is_private:
        if input.owner == out_backprop.owner:
            result = func_back_filter(input.inner_value, out_backprop.inner_value)
            result = PrivateTensorBase(inner_value=result,
                                       fixedpoint=input.fixedpoint + out_backprop.fixedpoint,
                                       owner=input.owner)
        else:
            result = BM_PrivateTensor_PrivateTensor(input, out_backprop, lambda_func, prf_flag=prf_flag)
    elif input_is_private and out_backprop_is_pair:
        result = BM_PrivateTensor_SharedPair(input, out_backprop, lambda_func, prf_flag=prf_flag)
    elif input_is_pair and out_backprop_is_private:
        result = BM_SharedPair_PrivateTensor(input, out_backprop, lambda_func, prf_flag=prf_flag)
    elif input_is_pair and out_backprop_is_pair:
        result = BM_SharedPair_SharedPair(input, out_backprop, lambda_func, prf_flag=prf_flag)
    else:
        raise Exception("type exception for type(x)={}, type(y)={}".format(type(input), type(out_backprop)))
    return result.dup_with_precision(fixed_point)


class Conv2dTriangle(BiliinearTriangle):
    def __init__(self, input_sizes, filter_sizes,
                   strides, padding,
                   data_format, dilations):
        model = StfConfig.conv_module
        def func_conv2d(x, y):
            return model.int64_conv2d(input=x, filter=y, strides=strides, padding=padding,
                                      data_format=data_format, dilations=dilations)

        def f_xy(x, y):
            return SharedTensorBase(func_conv2d(x.inner_value, y.inner_value))

        def func_back_input(f, ploss_pout):
            return model.int64_conv2d_backprop_input(input_sizes=input_sizes, filter=f,
                                                     out_backprop=ploss_pout, strides=strides,
                                                     padding=padding, data_format=data_format,
                                                     dilations=dilations)

        def f_yz(f, ploss_pout):
            return SharedTensorBase(func_back_input(f.inner_value, ploss_pout.inner_value))

        def func_back_filter(x, ploss_pout):
            return model.int64_conv2d_backprop_filter(input=x,
                                                      filter_sizes=filter_sizes,
                                                      out_backprop=ploss_pout,
                                                      strides=strides,
                                                      padding=padding,
                                                      data_format=data_format,
                                                      dilations=dilations)

        def f_zx(ploss_pout, x):
            return SharedTensorBase(func_back_filter(x.inner_value, ploss_pout.inner_value))

        super(Conv2dTriangle, self).__init__(f_xy, f_yz, f_zx)


#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : CNN
   Description :
   This is a 4-layer convolutional neural network that has the following structure.
   First is a 2-dimensional convolutional layer with 1 input channel, 16 output channels and a 5×5 filter.
   The activation functions following this layer are ReLU, followed by a 2×2 Average pool.
   The second layer is a 2-dimensional convolutional layer with 16 input channels, 16 output channels and another 5×5 filter.
   The activation functions following this layer are once again ReLU and a 2 × 2 Average pool.
   The third layer is an 256×100 fully-connected layer. The next activation function is ReLU.
   The final layer is a 100 × 10 linear layer and this is normalized using ASM(·) to get a probability distribution.
   Here, ASM(u_i)= ReLu(u_i)/sum(ReLu(u_0),ReLu(u_1),...ReLu(u_i),....)
   The loss function is cross entropy and stochastic gradient descent is used to minimize loss.
"""

from stensorflow.ml.nn.networks.NN import NN
from stensorflow.ml.nn.layers.layer import Layer
from stensorflow.ml.nn.layers.input import Input
from stensorflow.ml.nn.layers.relu import ReLU
from stensorflow.ml.nn.layers.conv2d import Conv2d
from stensorflow.ml.nn.layers.pooling import AveragePooling2D, MaxPooling2D
from stensorflow.basic.basic_class.pair import SharedVariablePair
from stensorflow.ml.nn.layers.loss import CrossEntropyLossWithSoftmax
from stensorflow.ml.nn.layers.flatten import Flatten
from stensorflow.ml.nn.layers.dense import Dense
import numpy as np
from stensorflow.random import random


class NetworkB(NN):
    def __init__(self, feature, label, pooling='max'):
        super(NetworkB, self).__init__()
        # input layer, init data；
        layer = Input(dim=28, x=feature)
        self.addLayer(layer)
        # convolutional layer with 1 input channel, 16 output channels and a 5×5 filter
        layer = Conv2d(output_dim=None, fathers=[layer], filters=16,
                       kernel_size=5, input_shape=[28, 28, 1])
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU(output_dim=layer.output_dim, fathers=[layer])
        self.addLayer(layer)
        # Average pool
        if pooling == 'max':
            layer = MaxPooling2D(output_dim=None, fathers=[layer], pool_size=(2, 2))
        else:
            layer = AveragePooling2D(output_dim=None, fathers=[layer], pool_size=(2, 2))

        self.addLayer(layer)
        # 16 input channels, 16 output channels and another 5×5 filter
        layer = Conv2d(output_dim=None, fathers=[layer], filters=16,
                       kernel_size=5, input_shape=layer.output_dim
                       )
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU(output_dim=layer.output_dim, fathers=[layer])
        self.addLayer(layer)
        # Average pool
        if pooling == 'max':
            layer = MaxPooling2D(output_dim=None, fathers=[layer], pool_size=(2, 2))
        else:
            layer = AveragePooling2D(output_dim=None, fathers=[layer], pool_size=(2, 2))

        self.addLayer(layer)
        # flatten data, only consider data_format = "NWHC"
        layer = Flatten(output_dim=None, fathers=[layer])
        self.addLayer(layer)
        # 全连接层
        # 256×100 fully-connected layer
        layer = Dense(output_dim=100, fathers=[layer])
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU(output_dim=layer.output_dim, fathers=[layer])
        self.addLayer(layer)
        # a 100 × 10 linear layer
        layer = Dense(output_dim=10, fathers=[layer])
        self.addLayer(layer)
        # 输出层
        layer_label = Input(dim=10, x=label)
        self.addLayer(ly=layer_label)
        # 损失层
        layer_loss = CrossEntropyLossWithSoftmax(layer_score=layer, layer_label=layer_label)
        self.addLayer(ly=layer_loss)

    def predict(self, x, out_prob=True):
        """
        Generates output predictions for the input samples.
        :param x: Input samples.
        :param out_prob: True or False
        :return: The predicted result or probability
        """
        self.cut_off()
        # Input Layer
        l_input = self.layers[0]
        assert isinstance(l_input, Input)
        l_input.replace(x)
        self.layers[0] = l_input
        # Output Layer
        ly = self.layers[-1]
        if not isinstance(ly, Layer):
            raise Exception("l must be a Layer")
        else:
            ly.forward()

        if out_prob:
            return ly.y
        else:
            return ly.score

    def predict_to_file(self, sess, x, predict_file_name,
                        pred_batch_num, model_file_machine,
                        with_sigmoid):
        """
        Save prediction results to file
        Computation is done in batches.
        :param sess: session
        :param x: PrivateTensor. Input samples.
        :param predict_file_name: String.
        :param pred_batch_num: Number of samples per batch.
        :param model_file_machine: String.
        :param with_sigmoid:
        :return:
        """
        y_pred = self.predict(x=x,  out_prob=with_sigmoid)
        id_y_pred = y_pred.to_tf_str(owner=model_file_machine)
        random.random_init(sess)
        # 分批写入文件
        with open(predict_file_name, "w") as f:
            for batch in range(pred_batch_num):
                records = sess.run(id_y_pred)
                records = "\n".join(records.astype('str'))
                # records.to_file()
                f.write(records + "\n")

    def print(self, model_file_machine, sess):
        """
        print model weights
        """
        for ly in self.layers:
            if isinstance(ly, Dense) or isinstance(ly, Conv2d):
                print(ly)
                for weight in ly.w:
                    print(weight)
                    weight = weight.to_tf_tensor(owner=model_file_machine)
                    print("***")
                    print(sess.run(weight))

    def replace_weight(self, keras_weight):
        """
        将传入的权重赋值给当前网络
        """
        i = 0
        for ly in self.layers:
            if isinstance(ly, Conv2d):
                # 用传入的权重直接进行预测
                # kernel = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i].shape)
                # kernel.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel
                # 用传入的权重进行训练
                kernel = ly.w[0]
                assert isinstance(kernel, SharedVariablePair)
                kernel.load_from_tf_tensor(keras_weight[i])
                i += 1

            if isinstance(ly, Dense):
                # 用传入的权重直接进行预测
                # kernel1 = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i].shape)
                # kernel1.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel1
                # kernel2 = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i+1].shape)
                # kernel2.load_from_numpy(keras_weight[i+1])
                # ly.w[1] = kernel2
                # 用传入的权重进行训练
                kernel1 = ly.w[0]
                assert isinstance(kernel1, SharedVariablePair)
                kernel1.load_from_tf_tensor(keras_weight[i])
                kernel2 = ly.w[1]
                assert isinstance(kernel2, SharedVariablePair)
                kernel2.load_from_tf_tensor(keras_weight[i+1])
                i += 2

    def save_model(self, sess, save_file_path, model_file_machine):
        """
        Save the model weights to npz file.
        :param sess: session
        :param save_file_path: String, path to save the model.
        :param model_file_machine: String, the machine name.
        :return:
        """
        res = []
        for ly in self.layers:
            if isinstance(ly, Dense) or isinstance(ly, Conv2d):
                for weight in ly.w:
                    weight = weight.to_tf_tensor(owner=model_file_machine)
                    weight = sess.run(weight)
                    res.append(weight)
        res = np.array(res)
        np.savez(save_file_path, weight=res)







#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : algebra
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-08-03 20:00
   Description : description what the main function of this file
"""

from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedTensorBase
from stensorflow.basic.basic_class.base import SharedPairBase
# from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.global_var import StfConfig
import tensorflow as tf
from stensorflow.exception.exception import StfCondException


def concat(x: list, axis, fixed_point=None):
    """

    :param x:
    :param axis:
    :return:
    """
    is_SharedTensorBase = list(map(lambda u: isinstance(u, SharedTensorBase), x))
    if all(is_SharedTensorBase):  # all x are SharedTensorBase
        if len(set(map(lambda u: u.module, x))) > 1:
            raise Exception("all x must have same module")
        module = x[0].module
        inner_value = tf.concat(list(map(lambda u: u.inner_value, x)), axis=axis)
        return SharedTensorBase(inner_value=inner_value, module=module)
    elif any(is_SharedTensorBase):
        raise Exception("must all x are SharedTensorBase or none of x is SharedTensorBase")

    if fixed_point is None:
        fixed_point = max(min(list(map(lambda u: u.fixedpoint, x))), StfConfig.default_fixed_point)
    x = list(map(lambda u: u.dup_with_precision(fixed_point), x))

    if all(list(map(lambda u: isinstance(u, PrivateTensorBase), x))):  # all x are PrivateTensorBase
        if len(set(map(lambda u: u.owner, x))) == 1:  # all x have same owner
            if len(set(map(lambda u: u.module, x))) > 1:
                raise Exception("all x must have same module")
            # fixed_point = min(list(map(lambda u: u.fixedpoint, x)))
            # x = list(map(lambda u: u.dup_with_precision(fixed_point), x))
            x_tf_tensor = list(map(lambda u: u.inner_value, x))
            y = tf.concat(x_tf_tensor, axis=axis)
            return PrivateTensorBase(owner=x[0].owner, fixedpoint=fixed_point, inner_value=y, module=x[0].module)
        else:
            owners = list(set(map(lambda u: u.owner, x)))
            if len(owners) > 2:
                raise StfCondException("len(owners)<=2", "owners={}".format(owners))

            x = list(
                map(
                    lambda u: u.to_SharedPairBase(
                        other_owner=owners[0] if owners[1] == u.owner else owners[1]),
                    x)
            )
            return concat(x, axis=axis, fixed_point=fixed_point)
    elif all(list(map(lambda u: isinstance(u, SharedPairBase), x))):  # all x[i] are SharedPairBase
        # y = x[0]
        # for i in range(1, len(x)):
        #     y = y.concat(x[i], axis=axis)
        # return y
        ownerL = x[0].ownerL
        ownerR = x[0].ownerR
        x = list(map(lambda u: u if u.ownerL == ownerL and u.ownerR == ownerR else u.mirror(), x))

        yL = concat(list(map(lambda u: u.xL, x)), axis=axis)
        yR = concat(list(map(lambda u: u.xR, x)), axis=axis)
        return SharedPairBase(ownerL=ownerL, ownerR=ownerR, xL=yL, xR=yR, fixedpoint=fixed_point)

    elif any(list(map(lambda u: isinstance(u, SharedPairBase), x))):  # exists x[i] is SharedPairBase
        for u in x:
            if isinstance(u, SharedPairBase):
                ownerL = u.ownerL
                ownerR = u.ownerR
                break
        for i in range(len(x)):
            if isinstance(x[i], PrivateTensorBase):
                if x[i].owner == ownerL:
                    other_owner = ownerR
                elif x[i].owner == ownerR:
                    other_owner = ownerL
                else:
                    raise Exception("must have x[i].owner == ownerL or x[i].owner == ownerR")
                x[i] = x[i].to_SharedPairBase(other_owner=other_owner)
        return concat(x, axis)
    else:
        raise Exception("x[i] must be PrivateTensorBase or SharedPairBase")



def stack(x: list, axis, fixed_point=None):
    """

    :param x:
    :param axis:
    :return:
    """
    is_SharedTensorBase = list(map(lambda u: isinstance(u, SharedTensorBase), x))
    if all(is_SharedTensorBase):  # all x are SharedTensorBase
        if len(set(map(lambda u: u.module, x))) > 1:
            raise Exception("all x must have same module")
        module = x[0].module
        inner_value = tf.stack(list(map(lambda u: u.inner_value, x)), axis=axis)
        return SharedTensorBase(inner_value=inner_value, module=module)
    elif any(is_SharedTensorBase):
        raise Exception("must all x are SharedTensorBase or none of x is SharedTensorBase")

    if fixed_point is None:
        fixed_point = max(min(list(map(lambda u: u.fixedpoint, x))), StfConfig.default_fixed_point)
    x = list(map(lambda u: u.dup_with_precision(fixed_point), x))

    if all(list(map(lambda u: isinstance(u, PrivateTensorBase), x))):  # all x are PrivateTensorBase
        if len(set(map(lambda u: u.owner, x))) == 1:  # all x have same owner
            if len(set(map(lambda u: u.module, x))) > 1:
                raise Exception("all x must have same module")
            # fixed_point = min(list(map(lambda u: u.fixedpoint, x)))
            # x = list(map(lambda u: u.dup_with_precision(fixed_point), x))
            x_tf_tensor = list(map(lambda u: u.inner_value, x))
            y = tf.stack(x_tf_tensor, axis=axis)
            return PrivateTensorBase(owner=x[0].owner, fixedpoint=fixed_point, inner_value=y, module=x[0].module)
        else:
            owners = list(set(map(lambda u: u.owner, x)))
            if len(owners) > 2:
                raise StfCondException("len(owners)<=2", "owners={}".format(owners))

            x = list(
                map(
                    lambda u: u.to_SharedPairBase(
                        other_owner=owners[0] if owners[1] == u.owner else owners[1]),
                    x)
            )
            return stack(x, axis=axis, fixed_point=fixed_point)
    elif all(list(map(lambda u: isinstance(u, SharedPairBase), x))):  # all x[i] are SharedPairBase
        # y = x[0]
        # for i in range(1, len(x)):
        #     y = y.concat(x[i], axis=axis)
        # return y
        ownerL = x[0].ownerL
        ownerR = x[0].ownerR
        x = list(map(lambda u: u if u.ownerL == ownerL and u.ownerR == ownerR else u.mirror(), x))

        yL = stack(list(map(lambda u: u.xL, x)), axis=axis)
        yR = stack(list(map(lambda u: u.xR, x)), axis=axis)
        return SharedPairBase(ownerL=ownerL, ownerR=ownerR, xL=yL, xR=yR, fixedpoint=fixed_point)

    elif any(list(map(lambda u: isinstance(u, SharedPairBase), x))):  # exists x[i] is SharedPairBase
        for u in x:
            if isinstance(u, SharedPairBase):
                ownerL = u.ownerL
                ownerR = u.ownerR
                break
        for i in range(len(x)):
            if isinstance(x[i], PrivateTensorBase):
                if x[i].owner == ownerL:
                    other_owner = ownerR
                elif x[i].owner == ownerR:
                    other_owner = ownerL
                else:
                    raise Exception("must have x[i].owner == ownerL or x[i].owner == ownerR")
                x[i] = x[i].to_SharedPairBase(other_owner=other_owner)
        return stack(x, axis)
    else:

        raise Exception("x[i] must be PrivateTensorBase or SharedPairBase")



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : logistic_regression
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-05-14 11:42
   Description : description what the main function of this file
"""

import tensorflow as tf
import stensorflow as stf
from typing import Union
import time
from stensorflow.basic.basic_class.pair import SharedPair, SharedVariablePair
from stensorflow.basic.basic_class.base import get_device
from stensorflow.basic.basic_class.private import PrivateTensor
import numpy as np
import os
from stensorflow.basic.operator.sigmoid import sigmoid_sin
from stensorflow.global_var import StfConfig
from stensorflow.homo_enc.homo_enc import homo_init
from stensorflow.random.random import random_init

class LogisticRegression2:
    """Contains methods to build and train logistic regression."""

    def __init__(self, num_features_L, num_features_R, learning_rate=0.01, l2_regularzation=1E-2):
        """
        :param num_features_L:  excluding label. part X -> L
        :param num_features_R:  including label. part Y -> R
        :param learning_rate:
        :param l2_regularzation:
        """
        # self.w = tfe.define_private_variable(
        #     tf.random_uniform([num_features, 1], -0.01, 0.01))
        assert (num_features_L >= 0 and num_features_R >= 0), "features should be non-negative"
        self.num_features_L = num_features_L
        self.num_features_R = num_features_R
        self.num_features = num_features_L + num_features_R
        # self.w = tfe.define_private_variable(
        #     tf.random_uniform([num_features, 1], -1.0/np.sqrt(num_features), 1.0/np.sqrt(num_features)))
        self.w_L = SharedVariablePair(ownerL="L", ownerR="R", shape=[num_features_L, 1])
        self.w_R = SharedVariablePair(ownerL="L", ownerR="R", shape=[num_features_R, 1])
        # self.w = SharedVariablePair(ownerL="L", ownerR="R")
        # self.w_L.load_from_tf_tensor(tf.compat.v1.random_uniform([num_features_L, 1],
        #                                                     -1.0/np.sqrt(num_features_L + 1),
        #                                                     1.0/np.sqrt(num_features_L + 1)))
        # self.w_R.load_from_tf_tensor(tf.compat.v1.random_uniform([num_features_R, 1],
        #                                                     -1.0 / np.sqrt(num_features_R + 1),
        #                                                     1.0 / np.sqrt(num_features_R + 1)))
        self.w_L.load_from_numpy(np.random.uniform(size=[num_features_L, 1],
                                                   low=-1.0 / np.sqrt(num_features_L + 1),
                                                   high=1.0 / np.sqrt(num_features_L + 1)))
        self.w_R.load_from_numpy(np.random.uniform(size=[num_features_R, 1],
                                                   low=-1.0 / np.sqrt(num_features_R + 1),
                                                   high=1.0 / np.sqrt(num_features_R + 1)))

        self.b = SharedVariablePair(ownerL="L", ownerR="R", shape=[1])
        # self.b.load_from_tf_tensor(tf.zeros([1]))
        self.b.load_from_numpy(np.zeros(shape=[1]))
        self.learning_rate = learning_rate
        self.l2_regularzation = l2_regularzation
        with tf.control_dependencies(StfConfig.log_op_list):
            with tf.device(self.w_L.ownerL):
                StfConfig.log_op_list.append(tf.print("initialize w_L"))
                StfConfig.log_op_list.append(tf.print("initialize w_R"))
                StfConfig.log_op_list.append(tf.print("initialize b"))
            with tf.device(self.w_L.ownerR):
                StfConfig.log_op_list.append(tf.print("initialize w_L"))
                StfConfig.log_op_list.append(tf.print("initialize w_R"))
                StfConfig.log_op_list.append(tf.print("initialize b"))

    @property
    def weights(self):
        """

        :return:
        """
        return self.w_L, self.w_R, self.b

    def forward(self,
                x_L: Union[PrivateTensor, SharedPair],
                x_R: Union[PrivateTensor, SharedPair],
                with_sigmoid=True):
        """

        :param x:
        :param with_sigmoid:
        :return:
        """

        m = x_L @ self.w_L + x_R @ self.w_R
        out = m + self.b

        with tf.control_dependencies(StfConfig.log_op_list):
            with tf.device(m.ownerL):
                StfConfig.log_op_list.append(tf.print("construct graph: forward"))
            with tf.device(m.ownerR):
                StfConfig.log_op_list.append(tf.print("construct graph: forward"))

        if with_sigmoid:
            y_hat = sigmoid_sin(out, M=16)
        else:
            y_hat = out
        return y_hat

    def backward(self,
                 x_L: Union[PrivateTensor, SharedPair],
                 x_R: Union[PrivateTensor, SharedPair],
                 dy: Union[PrivateTensor, SharedPair], learning_rate=0.01):
        """
        :param x_L:
        :param x_R:
        :param dy:
        :param learning_rate:
        :return:
        """

        batch_size = x_L.shape[0]
        with tf.name_scope("backward"):
            dw_L = (x_L.transpose() @ dy) / batch_size + self.l2_regularzation * self.w_L
            dw_R = (x_R.transpose() @ dy) / batch_size + self.l2_regularzation * self.w_R
            db = dy.reduce_sum(axis=0) / batch_size
            assign_ops = [
                self.w_L.assign(self.w_L - learning_rate * dw_L),
                self.w_R.assign(self.w_R - learning_rate * dw_R),
                self.b.assign(self.b - learning_rate * db),
            ]
            with tf.control_dependencies(StfConfig.log_op_list):
                with tf.device(dw_L.ownerL):
                    StfConfig.log_op_list.append(tf.print("construct graph: backward"))
                with tf.device(dw_L.ownerR):
                    StfConfig.log_op_list.append(tf.print("construct graph: backward"))
            return assign_ops

    def loss_grad(self, y: Union[PrivateTensor, SharedPair], y_hat: Union[PrivateTensor, SharedPair]):
        """

        :param y:
        :param y_hat:
        :return:
        """
        with tf.name_scope("loss-grad"):
            dy = y_hat - y
            return dy

    def fit_batch(self,
                  x_L: Union[PrivateTensor, SharedPair],
                  x_R: Union[PrivateTensor, SharedPair],
                  y: Union[PrivateTensor, SharedPair]):
        """

        :param x:
        :param y:
        :return:
        """
        with tf.name_scope("fit-batch"):
            y_hat = self.forward(x_L=x_L, x_R=x_R, with_sigmoid=True)
            dy = self.loss_grad(y, y_hat)
            fit_batch_op = self.backward(x_L, x_R, dy, self.learning_rate)
            return fit_batch_op

    def fit(self, sess,
            x_L: Union[PrivateTensor, SharedPair],
            x_R: Union[PrivateTensor, SharedPair],
            y: Union[PrivateTensor, SharedPair],
            num_batches):
        """

        :param sess:
        :param x_L: x part features
        :param x_R: y part features
        :param y:
        :param num_batches:
        :param progress_file:
        """
        fit_batch_op = self.fit_batch(x_L, x_R, y)
        print("StfConfig.log_op_list=", StfConfig.log_op_list)
        if StfConfig.pre_produce_flag and StfConfig.offline_model:
            pre_produce_op = tf.group(StfConfig.pre_produce_list)
        if sess is not None:
            sess.run(tf.compat.v1.global_variables_initializer())
            sess.run(StfConfig.log_op_list)
        start_time = time.time()
        if sess is not None:
            for batch in range(num_batches):
                print("batch", batch)
                if StfConfig.pre_produce_flag and StfConfig.offline_model:
                    sess.run(pre_produce_op)
                else:
                    sess.run(fit_batch_op)
        end_time = time.time()
        print("train time=", end_time - start_time)

    def predict_batch(self,
                      x_L: Union[PrivateTensor, SharedPair],
                      x_R: Union[PrivateTensor, SharedPair],
                      owner):
        """
        :param x_L:
        :param x_R:
        :return:
        """
        y_hat = self.forward(x_L=x_L, x_R=x_R, with_sigmoid=False)

        y_hat = y_hat.to_tf_tensor(owner=owner)

        # y_hat = y_hat / tf.cast(x.shape[1], dtype='float32')
        y_hat = tf.sigmoid(y_hat)
        return y_hat

    def predict_batch_simple(self,
                      x_L: Union[PrivateTensor, SharedPair],
                      x_R: Union[PrivateTensor, SharedPair],
                      owner):
        """
        :param x_L:
        :param x_R:
        :return:
        """

        w_L = self.w_L.to_tf_tensor(x_L.owner)
        w_R = self.w_R.to_tf_tensor(x_R.owner)

        with tf.device(x_L.owner):
            y_hat_L = x_L.to_tf_tensor(x_L.owner) @ w_L
        with tf.device(x_R.owner):
            y_hat_R = x_R.to_tf_tensor(x_R.owner) @ w_R
        with tf.device(get_device(owner)):
            y_hat = y_hat_L + y_hat_R + self.b.to_tf_tensor(owner)
        # y_hat = y_hat / tf.cast(x.shape[1], dtype='float32')
            y_hat = tf.sigmoid(y_hat)

        with tf.control_dependencies(StfConfig.log_op_list):
            with tf.device(x_L.owner):
                StfConfig.log_op_list.append(tf.print("construct graph: y_hat_L=x_L@w_L"))
            with tf.device(x_R.owner):
                StfConfig.log_op_list.append(tf.print("construct graph: y_hat_R=x_R@w_R"))
        with tf.control_dependencies(StfConfig.log_op_list):
            with tf.device(get_device(owner)):
                StfConfig.log_op_list.append(tf.print("construct graph: predict = y_hat_L + y_hat_R + b"))
        return y_hat

    # def predict(self, sess, x_L, x_R, file_name, num_batches, idx,
    #             progress_file, owner, record_num_ceil_mod_batch_size):
    #     """
    #
    #     :param sess:
    #     :param x_L:
    #     :param x_R:
    #     :param file_name:  write prediction
    #     :param num_batches:
    #     :param idx:
    #     :param progress_file:
    #     :param owner: output owner
    #     :param record_num_ceil_mod_batch_size:
    #     """
    #
    #     owner = get_device(owner)
    #     predict_batch = self.predict_batch(x_L=x_L, x_R=x_R, owner=owner)
    #
    #     with tf.device(owner):
    #         predict_batch = tf.strings.as_string(predict_batch)
    #
    #         predict_batch = tf.concat([idx, predict_batch], axis=1)
    #         predict_batch = tf.compat.v1.reduce_join(predict_batch, axis=1, separator=", ")
    #         # predict_batch=tf.reduce_join(predict_batch, separator="\n")
    #         if sess is not None:
    #             with open(file_name, "w") as f, open(progress_file, "a") as progress_file:
    #                 for batch in range(num_batches):
    #
    #                     records = sess.run(predict_batch)
    #
    #                     if batch == num_batches - 1:
    #                         records = records[0:record_num_ceil_mod_batch_size]
    #
    #                     # records = str(records, encoding="utf8")
    #                     records = "\n".join(records.astype('str'))
    #                     # records.to_file()
    #                     f.write(records + "\n")
    #
    #                     # if (batch % 10 == 0):
    #                     if batch % (1 + int(num_batches / 100)) == 0:
    #                         progress_file.write(str(1.0 * batch / num_batches) + "\n")
    #                         progress_file.flush()

    def predict(self, id, xL_test, xR_test, pred_batch_num, sess, predict_file=None):
        if predict_file is None:
            predict_file = StfConfig.predict_to_file
        if StfConfig.pre_produce_flag and StfConfig.offline_model:
            StfConfig.pre_produce_list = []
        y_pred = self.predict_batch(x_L=xL_test, x_R=xR_test, owner="R")
        if StfConfig.pre_produce_flag and StfConfig.offline_model:
            pre_produce_op = tf.group(StfConfig.pre_produce_list)
            homo_init(sess)
        random_init(sess)  # initialize the random model using in predict

        with open(predict_file, "w") as f:
            for batch in range(pred_batch_num):
                if StfConfig.pre_produce_flag and StfConfig.offline_model:
                    sess.run(pre_produce_op)
                else:
                    ids, records = sess.run([id, y_pred])
                    ids = ids.astype('str')
                    records = records.astype('str')
                    records = np.concatenate([ids, records], axis=1)
                    records = "\n".join([",".join(line) for line in records])
                    f.write(records + "\n")


    def predict_simple(self, id, xL_test, xR_test, pred_batch_num, sess, predict_file=None):
        StfConfig.log_op_list = []
        if predict_file is None:
            predict_file = StfConfig.predict_to_file
        if StfConfig.pre_produce_flag and StfConfig.offline_model:
            StfConfig.pre_produce_list = []
        y_pred = self.predict_batch_simple(x_L=xL_test, x_R=xR_test, owner="R")
        if StfConfig.pre_produce_flag and StfConfig.offline_model:
            pre_produce_op = tf.group(StfConfig.pre_produce_list)
            homo_init(sess)
        random_init(sess)  # initialize the random model using in predict
        sess.run(StfConfig.log_op_list)

        with open(predict_file, "w") as f:
            for batch in range(pred_batch_num):
                if StfConfig.pre_produce_flag and StfConfig.offline_model:
                    sess.run(pre_produce_op)
                else:
                    ids, records = sess.run([id, y_pred])
                    ids = ids.astype('str')
                    records = records.astype('str')
                    records = np.concatenate([ids, records], axis=1)
                    records = "\n".join([",".join(line) for line in records])
                    f.write(records + "\n")





    def save(self, model_file_path, model_file_machine="R"):

        if not os.path.exists(model_file_path):
            raise Exception("path not exist.{}".format(model_file_path))
        save_ops = []
        for i in range(len(self.weights)):
            model_file_path_i = os.path.join(model_file_path, "param_{i}".format(i=i))
            # save_op = tfe.define_output(modelFileMachine, [self.weights[i], modelFilePath_i], _save)
            save_op = self.weights[i].reshape([-1]).to_private(model_file_machine).to_file(
                path=model_file_path_i, separator=", ", precision=StfConfig.to_str_precision)
            save_ops = save_ops + [save_op]
        save_op = tf.group(*save_ops)
        return save_op

    def save_as_plaintext(self, model_file_path, model_file_machine="R"):
        private_b = self.b.to_private(model_file_machine)
        private_w_L = self.w_L.reshape([-1]).to_private(model_file_machine)
        private_w_R = self.w_R.reshape([-1]).to_private(model_file_machine)
        # CommonStf_config.default_logger.info("model_file_path:{}, shape, private_b:{}, w_L:{}, w_R:{}".format(
        #     model_file_path, private_b.shape, private_w_L.shape, private_w_R.shape))
        weights = private_b.concat(private_w_L, axis=0).concat(private_w_R, axis=0)
        save_op = weights.to_file(path=model_file_path, separator=", ", precision=StfConfig.to_str_precision)
        return save_op

    def load(self, model_file_path, model_file_machine="R"):
        self.w_L = stf.PrivateVariable(owner=model_file_machine)
        self.w_R = stf.PrivateVariable(owner=model_file_machine)
        self.b = stf.PrivateVariable(owner=model_file_machine)

        self.w_L.load_first_line_from_file(os.path.join(model_file_path, "param_0"), col_num=self.num_features_L)
        self.w_L = self.w_L.reshape([self.num_features_L, 1])

        self.w_R.load_first_line_from_file(os.path.join(model_file_path, "param_1"), col_num=self.num_features_R)
        self.w_R = self.w_R.reshape([self.num_features_R, 1])

        self.b.load_first_line_from_file(os.path.join(model_file_path, "param_2"), col_num=1)
        self.b = self.b.reshape([1, 1])



#!/usr/bin/env python
# coding=utf-8
"""
   Alipay.com Inc.
   Copyright (c) 2004-2022 All Rights Reserved.
   ------------------------------------------------------
   File Name : read_libsvm
   Author : qizhi.zqz
   Email: qizhi.zqz@alibaba-inc.com
   Create Time : 2022/3/22 下午3:10
   Description : description what the main function of this file
"""
import tensorflow as tf

def decode_libsvm(line):
    columns = tf.compat.v1.string_split([line], ' ')
    labels = tf.compat.v1.string_to_number(columns.values[0], out_type=tf.float32)
    splits = tf.compat.v1.string_split(columns.values[1:], ':')
    print("splits=",splits)
    """
    SparseTensorValue(indices=array([[ 0,  0],
       [ 0,  1],
       [ 1,  0],
       [ 1,  1],
       [ 2,  0],
       [ 2,  1],
       [ 3,  0],
       [ 3,  1],
       [ 4,  0],
       [ 4,  1],
       [ 5,  0],
       [ 5,  1],
       [ 6,  0],
       [ 6,  1],
       [ 7,  0],
       [ 7,  1],
       [ 8,  0],
       [ 8,  1],
       [ 9,  0],
       [ 9,  1],
       [10,  0],
       [10,  1],
       [11,  0],
       [11,  1],
       [12,  0],
       [12,  1],
       [13,  0],
       [13,  1]]), values=array([b'3', b'1', b'11', b'1', b'14', b'1', b'19', b'1', b'39', b'1',
       b'42', b'1', b'55', b'1', b'64', b'1', b'67', b'1', b'73', b'1',
       b'75', b'1', b'76', b'1', b'80', b'1', b'83', b'1'], dtype=object), dense_shape=array([14,  2]))
    """
    id_vals = tf.reshape(splits.values,splits.dense_shape)
    print("id_vals=", id_vals)
    """
    [[b'3' b'1']
 [b'11' b'1']
 [b'14' b'1']
 [b'19' b'1']
 [b'39' b'1']
 [b'42' b'1']
 [b'55' b'1']
 [b'64' b'1']
 [b'67' b'1']
 [b'73' b'1']
 [b'75' b'1']
 [b'76' b'1']
 [b'80' b'1']
 [b'83' b'1']]
    """
    feat_ids, feat_vals = tf.split(id_vals,num_or_size_splits=2,axis=1)
    feat_ids = tf.compat.v1.string_to_number(feat_ids, out_type=tf.int32)
    feat_ids = tf.squeeze(feat_ids, axis=-1)
    feat_vals = tf.compat.v1.string_to_number(feat_vals, out_type=tf.float32)
    feat_vals = tf.squeeze(feat_vals, axis=-1)
    return {"feat_ids": feat_ids, "feat_vals": feat_vals}, labels
    #return id_vals



def decode_libsvm_x(line, field_delim=" ", skip_col_num=0):
    columns = tf.compat.v1.string_split([line], field_delim)
    #labels = tf.compat.v1.string_to_number(columns.values[0], out_type=tf.float32)
    splits = tf.compat.v1.string_split(columns.values[skip_col_num:], ':')
    id_vals = tf.reshape(splits.values,splits.dense_shape)
    feat_ids, feat_vals = tf.split(id_vals, num_or_size_splits=2,axis=1)
    feat_ids = tf.compat.v1.string_to_number(feat_ids, out_type=tf.int64)
    # feat_ids = tf.squeeze(feat_ids, axis=-1)
    feat_vals = tf.compat.v1.string_to_number(feat_vals, out_type=tf.float32)
    feat_vals = tf.squeeze(feat_vals, axis=-1)
    return feat_ids, feat_vals





# Extract lines from input files using the Dataset API, can pass one filename or filename list

def load_libsvm_x_to_dense(filenames, feature_num, batch_size, field_delim=" ", skip_row_num=1, skip_col_num=0,
                           repeat=1, clip_value=None, scale=1.0, map_fn=None, output_col_num=None, buffer_size=0):
    def decode_libsvm_x_to_dense(line):
        feat_ids, feat_vals = decode_libsvm_x(line, field_delim=field_delim, skip_col_num=skip_col_num)
        r_sparse = tf.SparseTensor(indices=feat_ids-1, values=feat_vals, dense_shape=[feature_num])
        return tf.compat.v1.sparse_tensor_to_dense(r_sparse, default_value=0.0)
    def clip(r):
        if clip_value is None:
            return r * scale if scale != 1.0 else r
        else:
            return tf.clip_by_value(r * scale, -clip_value, clip_value)
    if output_col_num is None:
        output_col_num = feature_num - skip_col_num
    dataset = tf.compat.v1.data.TextLineDataset(filenames, buffer_size=buffer_size)\
        .skip(skip_row_num).map(decode_libsvm_x_to_dense, num_parallel_calls=batch_size)\
        .prefetch(1000)

    data = dataset.repeat(repeat).batch(batch_size=batch_size).make_one_shot_iterator().get_next()
    # print("feature_num=", feature_num)
    # print("output_col_num=", output_col_num)
    data = tf.reshape(data, [batch_size, output_col_num])
    if map_fn is not None:
        data = data.map(map_func=map_fn)
    data = clip(data)
    return data

def load_libsvm_x_to_sparse(filenames, batch_size, feature_num, skip_row_num=0):
    def decode_libsvm_x_to_sparse(line):
        feat_ids, feat_vals = decode_libsvm_x(line)
        r_sparse = tf.SparseTensor(indices=feat_ids-1, values=feat_vals, dense_shape=[feature_num])
        return r_sparse
    dataset = tf.compat.v1.data.TextLineDataset(filenames).skip(skip_row_num).map(decode_libsvm_x_to_sparse, num_parallel_calls=batch_size).prefetch(1000)
    data = dataset.batch(batch_size=batch_size).make_one_shot_iterator().get_next()
    return data

def load_libsvm_x(filenames, feature_num, batch_size, field_delim=" ", skip_row_num=1, skip_col_num=0,
                           repeat=1, clip_value=None, scale=1.0, map_fn=None,
                  output_col_num=None, buffer_size=0, sparse_flag=False):
    if not sparse_flag:
        return load_libsvm_x_to_dense(filenames, feature_num, batch_size, field_delim, skip_row_num, skip_col_num,
                           repeat, clip_value, scale, map_fn, output_col_num, buffer_size)
    else:
        return load_libsvm_x_to_sparse(filenames, batch_size, feature_num)


#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2021 All Rights Reserved.
   ------------------------------------------------------
   File Name : random.py
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2021/5/24 下午8:34
   Description : description what the main function of this file
"""
import tensorflow as tf
from stensorflow.global_var import StfConfig
from stensorflow.exception.exception import StfTypeException, StfCondException, check_is_not_None


def gen_rint64(shape):
    random_mod = StfConfig.random_module
    check_is_not_None(shape=shape)
    step = tf.Variable(initial_value=[0], trainable=False)
    tf.compat.v1.add_to_collection(StfConfig.coll_name_vars_random, step)
    x = random_mod.rint64(shape=shape, step=step)
    return x


def gen_ZZn(shape, module):
    random_mod = StfConfig.random_module
    check_is_not_None(shape=shape)
    step = tf.Variable(initial_value=[0], trainable=False)
    tf.compat.v1.add_to_collection(StfConfig.coll_name_vars_random, step)
    x = random_mod.rint64(shape=shape, step=step)
    return x % module


def get_seed():
    random_mod = StfConfig.random_module
    seed = random_mod.get_seed(0)
    return seed


def gen_rint64_from_seed(shape, seed):
    random_mod = StfConfig.random_module
    check_is_not_None(seed=seed)
    step = tf.Variable(initial_value=[0], dtype='int64', trainable=False)
    tf.compat.v1.add_to_collection(StfConfig.coll_name_vars_random, step)
    x = random_mod.rint64_from_seed(shape=shape, seed=seed + step)  # Only use seed in first time in this OP
    return x


def gen_ZZn_from_seed(shape, seed, module: int):
    if not isinstance(module, int):
        raise StfTypeException("module", 'int', type(module))
    elif module <= 0:
        raise StfCondException("module>0", "module={}".format(module))
    random_mod = StfConfig.random_module
    check_is_not_None(seed=seed)
    step = tf.Variable(initial_value=[0], dtype='int64', trainable=False)
    tf.compat.v1.add_to_collection(StfConfig.coll_name_vars_random, step)
    x = random_mod.rint64_from_seed(shape=shape, seed=seed + step) % module  # Only use seed in first time in this OP
    return x


def random_init(sess=None):
    init_coll = tf.compat.v1.get_collection(StfConfig.coll_name_vars_random)
    init_random_op = tf.compat.v1.initialize_variables(init_coll)
    if sess is not None:
        sess.run(init_random_op)
    else:
        return init_random_op



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : module_transform
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-06-08 19:35
   Description : modular transform: https://eprint.iacr.org/2021/857
"""

from stensorflow.basic.basic_class.base import SharedPairBase, SharedTensorBase
from stensorflow.random.random import get_seed
from stensorflow.global_var import StfConfig
import tensorflow as tf


def module_transform_withoutPRF(a: SharedPairBase, new_module: int, compress_flag=None):
    if a.fixedpoint > 0:
        raise Exception("must have a.fixedpoint==0")
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    old_module = a.xL.module
    if old_module == 2:
        with tf.device(StfConfig.RS[0]):
            uL = a.to_SharedTensor_like().random_uniform_adjoint()
            uR = a.to_SharedTensor_like().random_uniform_adjoint()
            u = uL + uR
            u.module = new_module
            bL = u.random_uniform_adjoint()
            bR = u - bL

            if compress_flag:
                uL_compress = uL.to_compress_tensor()
                uR_compress = uR.to_compress_tensor()
                bL_compress = bL.to_compress_tensor()
                bR_compress = bR.to_compress_tensor()

        with tf.device(a.ownerL):
            if compress_flag:
                uL.decompress_from(uL_compress, uL.shape)
                bL.decompress_from(bL_compress, bL.shape)
            zL = a.xL - uL
            if compress_flag:
                zL_compress = zL.to_compress_tensor()
        with tf.device(a.ownerR):
            if compress_flag:
                uR.decompress_from(uR_compress, uR.shape)
                bR.decompress_from(bR_compress, bR.shape)
            zR = a.xR - uR
            if compress_flag:
                zR_compress = zR.to_compress_tensor()
                zL.decompress_from(zL_compress, zL.shape)
            z = zL + zR
            yR = tf.where(tf.cast(z.inner_value, 'bool'), x=-bR.inner_value, y=bR.inner_value)
            yR = SharedTensorBase(inner_value=yR + z.inner_value, module=new_module)
        with tf.device(a.ownerL):
            if compress_flag:
                zR.decompress_from(zR_compress, zR.shape)
            z = zL + zR
            yL = tf.where(tf.cast(z.inner_value, 'bool'), x=-bL.inner_value, y=bL.inner_value)
            yL = SharedTensorBase(inner_value=yL, module=new_module)

        return SharedPairBase(ownerL=a.ownerL, ownerR=a.ownerR, xL=yL, xR=yR, fixedpoint=0)
    else:
        raise NotImplementedError


def module_transform_withPRF(a: SharedPairBase, new_module: int, compress_flag=None):
    if a.fixedpoint > 0:
        raise Exception("must have a.fixedpoint==0")
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    old_module = a.xL.module
    if old_module == 2:
        with tf.device(StfConfig.RS[0]):
            seed_aL = get_seed()
            seed_aR = get_seed()
            seed_b = get_seed()
            aL_adjoint = a.to_SharedTensor_like().random_uniform_adjoint(seed_aL)
            aR_adjoint = a.to_SharedTensor_like().random_uniform_adjoint(seed_aR)
            b = aL_adjoint + aR_adjoint
            b.module = new_module
            bL = b.random_uniform_adjoint(seed_b)
            bR = b - bL
            if compress_flag:
                bR_compress = bR.to_compress_tensor()
        with tf.device(a.ownerL):
            aL_adjoint_onL = a.to_SharedTensor_like().random_uniform_adjoint(seed_aL)
            delta_aL = a.xL - aL_adjoint_onL
            if compress_flag:
                delta_aL_compress = delta_aL.to_compress_tensor()
        with tf.device(a.ownerR):
            aR_adjoint_onR = a.to_SharedTensor_like().random_uniform_adjoint(seed_aR)
            delta_aR = a.xR - aR_adjoint_onR
            if compress_flag:
                bR.decompress_from(bR_compress, bR.shape)
                delta_aR_compress = delta_aR.to_compress_tensor()
                delta_aL_onR = delta_aL.decompress_from_to_new(delta_aL_compress)
            else:
                delta_aL_onR = delta_aL.identity()
            delta_a_onR = delta_aL_onR + delta_aR
            yR = tf.where(tf.cast(delta_a_onR.inner_value, 'bool'), x=-bR.inner_value, y=bR.inner_value)
            yR = SharedTensorBase(inner_value=yR + delta_a_onR.inner_value, module=new_module)
        with tf.device(a.ownerL):
            if compress_flag:
                delta_aR_onL = delta_aR.decompress_from_to_new(delta_aR_compress)
            else:
                delta_aR_onL = delta_aR.identity()
            delta_a_onL = delta_aL + delta_aR_onL
            yL = tf.where(tf.cast(delta_a_onL.inner_value, 'bool'), x=-bL.inner_value, y=bL.inner_value)
            yL = SharedTensorBase(inner_value=yL, module=new_module)
        return SharedPairBase(ownerL=a.ownerL, ownerR=a.ownerR, xL=yL, xR=yR, fixedpoint=0)
    else:
        raise NotImplementedError


def module_transform(a: SharedPairBase, new_module: int, prf_flag=None, compress_flag=None):
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    if prf_flag:
        return module_transform_withPRF(a, new_module, compress_flag)
    else:
        return module_transform_withoutPRF(a, new_module, compress_flag)
